/*
 * Copyright (C) 2005-2006 Atmel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <common.h>

/*
 * The u-boot networking stack is a little weird.  It seems like the
 * networking core allocates receive buffers up front without any
 * regard to the hardware that's supposed to actually receive those
 * packets.
 *
 * The MACB receives packets into 128-byte receive buffers, so the
 * buffers allocated by the core isn't very practical to use.  We'll
 * allocate our own, but we need one such buffer in case a packet
 * wraps around the DMA ring so that we have to copy it.
 *
 * Therefore, define CONFIG_SYS_RX_ETH_BUFFER to 1 in the board-specific
 * configuration header.  This way, the core allocates one RX buffer
 * and one TX buffer, each of which can hold a ethernet packet of
 * maximum size.
 *
 * For some reason, the networking core unconditionally specifies a
 * 32-byte packet "alignment" (which really should be called
 * "padding").  MACB shouldn't need that, but we'll refrain from any
 * core modifications here...
 */

#include <net.h>
#include <netdev.h>
#include <malloc.h>
#include <miiphy.h>

#include <linux/mii.h>
#include <asm/io.h>
#include <asm/dma-mapping.h>
#include <asm/arch/clk.h>

#include "macb.h"

#define barrier() asm volatile("" ::: "memory")

#define CONFIG_SYS_MACB_RX_BUFFER_SIZE          4096
#define CONFIG_SYS_MACB_RX_RING_SIZE            (CONFIG_SYS_MACB_RX_BUFFER_SIZE / 128)
#define CONFIG_SYS_MACB_TX_RING_SIZE            16
#define CONFIG_SYS_MACB_TX_TIMEOUT              1000
#define CONFIG_SYS_MACB_AUTONEG_TIMEOUT         5000000

struct macb_dma_desc
{
  u32   addr;
  u32   ctrl;
};

#define RXADDR_USED             0x00000001
#define RXADDR_WRAP             0x00000002

#define RXBUF_FRMLEN_MASK       0x00000fff
#define RXBUF_FRAME_START       0x00004000
#define RXBUF_FRAME_END         0x00008000
#define RXBUF_TYPEID_MATCH      0x00400000
#define RXBUF_ADDR4_MATCH       0x00800000
#define RXBUF_ADDR3_MATCH       0x01000000
#define RXBUF_ADDR2_MATCH       0x02000000
#define RXBUF_ADDR1_MATCH       0x04000000
#define RXBUF_BROADCAST         0x80000000

#define TXBUF_FRMLEN_MASK       0x000007ff
#define TXBUF_FRAME_END         0x00008000
#define TXBUF_NOCRC             0x00010000
#define TXBUF_EXHAUSTED         0x08000000
#define TXBUF_UNDERRUN          0x10000000
#define TXBUF_MAXRETRY          0x20000000
#define TXBUF_WRAP              0x40000000
#define TXBUF_USED              0x80000000

struct macb_device
{
  void                  *regs;

  unsigned int          rx_tail;
  unsigned int          tx_head;
  unsigned int          tx_tail;

  void                  *rx_buffer;
  void                  *tx_buffer;
  struct macb_dma_desc  *rx_ring;
  struct macb_dma_desc  *tx_ring;

  unsigned long         rx_buffer_dma;
  unsigned long         rx_ring_dma;
  unsigned long         tx_ring_dma;

  const struct device   *dev;
  struct eth_device     netdev;
  unsigned short        phy_addr;

  unsigned              rockwell_switch_present;
  unsigned              rockwell_switch_running;
};
#define to_macb(_nd) container_of(_nd, struct macb_device, netdev)

struct macb_device *macbp;

static void macb_mdio_write(struct macb_device *macb, u8 reg, u16 value)
{
  unsigned long netctl;
  unsigned long netstat;
  unsigned long frame;

  netctl = macb_readl(macb, NCR);
  netctl |= MACB_BIT(MPE);
  macb_writel(macb, NCR, netctl);

  frame = (MACB_BF(SOF, 1)
           | MACB_BF(RW, 1)
           | MACB_BF(PHYA, macb->phy_addr)
           | MACB_BF(REGA, reg)
           | MACB_BF(CODE, 2)
           | MACB_BF(DATA, value));
  macb_writel(macb, MAN, frame);

  do
    {
      netstat = macb_readl(macb, NSR);
    }
  while (!(netstat & MACB_BIT(IDLE)));

  netctl = macb_readl(macb, NCR);
  netctl &= ~MACB_BIT(MPE);
  macb_writel(macb, NCR, netctl);
}

static u16 macb_mdio_read(struct macb_device *macb, u8 reg)
{
  unsigned long netctl;
  unsigned long netstat;
  unsigned long frame;

  netctl = macb_readl(macb, NCR);
  netctl |= MACB_BIT(MPE);
  macb_writel(macb, NCR, netctl);

  frame = (MACB_BF(SOF, 1)
           | MACB_BF(RW, 2)
           | MACB_BF(PHYA, macb->phy_addr)
           | MACB_BF(REGA, reg)
           | MACB_BF(CODE, 2));
  macb_writel(macb, MAN, frame);

  do
    {
      netstat = macb_readl(macb, NSR);
    }
  while (!(netstat & MACB_BIT(IDLE)));

  frame = macb_readl(macb, MAN);

  netctl = macb_readl(macb, NCR);
  netctl &= ~MACB_BIT(MPE);
  macb_writel(macb, NCR, netctl);

  return MACB_BFEXT(DATA, frame);
}

#if defined(CONFIG_CMD_MII)

int macb_miiphy_read(char *devname, u8 phy_adr, u8 reg, u16 *value)
{
  struct eth_device *dev = eth_get_dev_by_name(devname);
  struct macb_device *macb = to_macb(dev);

  if ( macb->phy_addr != phy_adr )
    return -1;

  *value = macb_mdio_read(macb, reg);

  return 0;
}

int macb_miiphy_write(char *devname, u8 phy_adr, u8 reg, u16 value)
{
  struct eth_device *dev = eth_get_dev_by_name(devname);
  struct macb_device *macb = to_macb(dev);

  if ( macb->phy_addr != phy_adr )
    return -1;

  macb_mdio_write(macb, reg, value);

  return 0;
}
#endif


#if defined(CONFIG_CMD_NET)

static int macb_send(struct eth_device *netdev, volatile void *packet,
                     int length)
{
  struct macb_device *macb = to_macb(netdev);
  unsigned long paddr, ctrl;
  unsigned int tx_head = macb->tx_head;
  int i;

  paddr = dma_map_single(packet, length, DMA_TO_DEVICE);

  ctrl = length & TXBUF_FRMLEN_MASK;
  ctrl |= TXBUF_FRAME_END;
  if (tx_head == (CONFIG_SYS_MACB_TX_RING_SIZE - 1))
    {
      ctrl |= TXBUF_WRAP;
      macb->tx_head = 0;
    }
  else
    macb->tx_head++;

  macb->tx_ring[tx_head].ctrl = ctrl;
  macb->tx_ring[tx_head].addr = paddr;
  barrier();
  macb_writel(macb, NCR, MACB_BIT(TE) | MACB_BIT(RE) | MACB_BIT(TSTART));

  /*
   * I guess this is necessary because the networking core may
   * re-use the transmit buffer as soon as we return...
   */
  for (i = 0; i <= CONFIG_SYS_MACB_TX_TIMEOUT; i++)
    {
      barrier();
      ctrl = macb->tx_ring[tx_head].ctrl;
      if (ctrl & TXBUF_USED)
        break;
      udelay(1);
    }

  dma_unmap_single(packet, length, paddr);

  if (i <= CONFIG_SYS_MACB_TX_TIMEOUT)
    {
      if (ctrl & TXBUF_UNDERRUN)
        printf("%s: TX underrun\n", netdev->name);
      if (ctrl & TXBUF_EXHAUSTED)
        printf("%s: TX buffers exhausted in mid frame\n",
               netdev->name);
    }
  else
    {
      printf("%s: TX timeout\n", netdev->name);
    }

  /* No one cares anyway */
  return 0;
}

static void reclaim_rx_buffers(struct macb_device *macb,
                               unsigned int new_tail)
{
  unsigned int i;

  i = macb->rx_tail;
  while (i > new_tail)
    {
      macb->rx_ring[i].addr &= ~RXADDR_USED;
      i++;
      if (i > CONFIG_SYS_MACB_RX_RING_SIZE)
        i = 0;
    }

  while (i < new_tail)
    {
      macb->rx_ring[i].addr &= ~RXADDR_USED;
      i++;
    }

  barrier();
  macb->rx_tail = new_tail;
}

static int macb_recv(struct eth_device *netdev)
{
  struct macb_device *macb = to_macb(netdev);
  unsigned int rx_tail = macb->rx_tail;
  void *buffer;
  int length;
  int wrapped = 0;
  u32 status;

  for (;;)
    {
      if (!(macb->rx_ring[rx_tail].addr & RXADDR_USED))
        return -1;

      status = macb->rx_ring[rx_tail].ctrl;
      if (status & RXBUF_FRAME_START)
        {
          if (rx_tail != macb->rx_tail)
            reclaim_rx_buffers(macb, rx_tail);
          wrapped = 0;
        }

      if (status & RXBUF_FRAME_END)
        {
          buffer = macb->rx_buffer + 128 * macb->rx_tail;
          length = status & RXBUF_FRMLEN_MASK;
          if (wrapped)
            {
              unsigned int headlen, taillen;

              headlen = 128 * (CONFIG_SYS_MACB_RX_RING_SIZE
                               - macb->rx_tail);
              taillen = length - headlen;
              memcpy((void *)NetRxPackets[0],
                     buffer, headlen);
              memcpy((void *)NetRxPackets[0] + headlen,
                     macb->rx_buffer, taillen);
              buffer = (void *)NetRxPackets[0];
            }

          NetReceive(buffer, length);
          if (++rx_tail >= CONFIG_SYS_MACB_RX_RING_SIZE)
            rx_tail = 0;
          reclaim_rx_buffers(macb, rx_tail);
        }
      else
        {
          if (++rx_tail >= CONFIG_SYS_MACB_RX_RING_SIZE)
            {
              wrapped = 1;
              rx_tail = 0;
            }
        }
      barrier();
    }

  return 0;
}

static void macb_phy_reset(struct macb_device *macb)
{
  struct eth_device *netdev = &macb->netdev;
  int i;
  u16 status, adv;

  adv = ADVERTISE_CSMA | ADVERTISE_ALL;
  macb_mdio_write(macb, MII_ADVERTISE, adv);
  printf("%s: Starting autonegotiation...\n", netdev->name);
  macb_mdio_write(macb, MII_BMCR, (BMCR_ANENABLE
                                   | BMCR_ANRESTART));

  for (i = 0; i < CONFIG_SYS_MACB_AUTONEG_TIMEOUT / 100; i++)
    {
      status = macb_mdio_read(macb, MII_BMSR);
      if (status & BMSR_ANEGCOMPLETE)
        break;
      udelay(100);
    }

  if (status & BMSR_ANEGCOMPLETE)
    printf("%s: Autonegotiation complete\n", netdev->name);
  else
    printf("%s: Autonegotiation timed out (status=0x%04x)\n",
           netdev->name, status);
}

#ifdef CONFIG_MACB_SEARCH_PHY
static int macb_phy_find(struct macb_device *macb)
{
  int i;
  u16 phy_id;

  /* Search for PHY... */
  for (i = 0; i < 32; i++)
    {
      macb->phy_addr = i;
      phy_id = macb_mdio_read(macb, MII_PHYSID1);
      if (phy_id != 0xffff)
        {
          printf("%s: PHY present at %d\n", macb->netdev.name, i);
          return 1;
        }
    }

  /* PHY isn't up to snuff */
  printf("%s: PHY not found", macb->netdev.name);

  return 0;
}
#endif /* CONFIG_MACB_SEARCH_PHY */

#define Bit(n) (1<<(n))

// mdio read/write routines that allow access to any address on the
// mdio bus.

static u16 mdio_read(struct macb_device *macb, u16 addr, u16 reg)
{
  unsigned old = macb->phy_addr = addr;
  u16 r = macb_mdio_read(macb, reg);
  macb->phy_addr = old;
  return r;
}

static void mdio_write(struct macb_device *macb, u16 addr, u16 reg, u16 val)
{
  unsigned old = macb->phy_addr = addr;
  macb_mdio_write(macb, reg, val);
  macb->phy_addr = old;
}

static int do_phyread(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  unsigned addr=1, reg=0, count=32, i;

  if (!macbp)
    {
      printf("macb driver not initialized\n");
      return 1;
    }

  if (argc > 4)
    {
      printf("ignoring extra arguments starting with '%s'\n",argv[4]);
      return 1;
    }

  if (argc > 1)
    {
      addr = simple_strtol(argv[1],NULL,16);
      if (argc > 2)
        {
          reg = simple_strtol(argv[2],NULL,16);
          count = 1;
          if (argc > 2)
            count = simple_strtol(argv[3],NULL,16);
        }
    }

  for (i=0; i<count; ++i)
    {
      if (i%8 == 0)
        {
          if (i) printf("\n");
          printf("%04x:", reg+i);
        }
      printf(" %04x", mdio_read(macbp, addr, reg+i));
    }
  printf("\n");
  return 0;
}
U_BOOT_CMD(phyrd, 4, 0, do_phyread, "Read PHY registers", "<phy-addr> <phy-reg> [count]");


static int do_phywrite(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  unsigned addr,reg,val;

  if (!macbp)
    {
      printf("macb driver not initialized\n");
      return 1;
    }

  if (argc != 4)
    {
      printf("usage: phywr <phy-addr> <phy-reg> <value>\n");
      return 1;
    }

  addr = simple_strtol(argv[1],NULL,16);
  reg  = simple_strtol(argv[2],NULL,16);
  val  = simple_strtol(argv[3],NULL,16);

  mdio_write(macbp, addr, reg, val);
  return 0;
}
U_BOOT_CMD(phywr, 4, 0, do_phywrite, "Write PHY register", "<phy-addr> <phy-reg> <value>");


// probe for Marvell switch and do needed initialization if found.

static int probe_M88E6095(struct macb_device *macb)
{
  int i;
  u16 u;
  struct eth_device *netdev = &macb->netdev;
  static const struct probe
  {
    u16 addr;
    u16 reg;
    u16 val;
  } probes[] = {{0x10,  3, 0x0953},
    {0x1a,  3, 0x0953},
    {0x1b, 18, 0x5555},
    {0x1b, 20, 0xaaaa},
    {0x1b, 25, 0x8100}
  };

  for (i=0; i< ((sizeof probes)/(sizeof probes[0])); ++i)
    {
      const struct probe *p = probes+i;
      if (p->val != mdio_read(macb, p->addr, p->reg))
        return 0;
    }

  printf("%s: detected Marvel 88E6095v3 switch\n",netdev->name);

  // make sure link between CPU and switch is in "up" state
  if ((mdio_read(macb, 0x1a, 1) & 0x00ff) != 0x00fd)
    {
      printf("%s: forcing link state=up on port 11\n",netdev->name);
      mdio_write(macb, 0x1a, 1, 0x00fd);
    }

  // disable PPU so we can talk to external PHYs
  u = mdio_read(macb, 0x1b, 4);
  mdio_write(macb, 0x1b, 4, u & ~Bit(14));

  // wait for PPU disable state
  while (mdio_read(macb, 0x1b, 0) & Bit(14))
    udelay(1000);

  // check for Marvell gigabit PHY on port 8, and disable gigabit
  // operation (our board's magnetics don't support it).

  if ((mdio_read(macb, 8, 2) == 0x0141) && ((mdio_read(macb, 8, 3) & 0xfff0) == 0x0c90))
    {
      printf("%s: detected Marvel 88E1112 PHY on port 8.\n",netdev->name);
      if (mdio_read(macb, 8, 9) & (Bit(8)+Bit(9)))
        {
          printf("%s: disabling gigabit negotiation on port 8.\n",netdev->name);
          u = mdio_read(macb, 8, 9);
          mdio_write(macb, 8, 9, u & ~(Bit(8)+Bit(9)));
          // force autonegotiation restart
          u = mdio_read(macb, 8, 0);
          mdio_write(macb, 8, 0, u | Bit(9));
        }
    }

  // re-enable PPU
  u = mdio_read(macb, 0x1b, 4);
  mdio_write(macb, 0x1b, 4, u | Bit(14));

  // I don't know why, but the switch won't forward traffic to the two
  // external PHYs if you don't do a read after diabling and
  // re-enabling PPU.  It doesn't seem to matter _what_ you read, but
  // without it, you get no traffic to/from the external PHYs
  mdio_read(macb, 0x1b, 0);

  return 1;
}

#define NoPhyPresent 1

static int macb_phy_init(struct macb_device *macb)
{
  struct eth_device *netdev = &macb->netdev;
  u32 ncfgr;
  u16 phy_id, status, adv, lpa;
  int media, speed, duplex;
  int i;

  probe_M88E6095(macb);

  if (NoPhyPresent)
    {
      // force to 100 full
      ncfgr = macb_readl(macb, NCFGR);
      ncfgr |= MACB_BIT(SPD) | MACB_BIT(FD);
      macb_writel(macb, NCFGR, ncfgr);
      printf("%s: forced 100Mbps full-duplex\n", netdev->name);
      return 1;
    }

#ifdef CONFIG_MACB_SEARCH_PHY
  /* Auto-detect phy_addr */
  if (!macb_phy_find(macb))
    {
      return 0;
    }
#endif /* CONFIG_MACB_SEARCH_PHY */

  /* Check if the PHY is up to snuff... */
  phy_id = macb_mdio_read(macb, MII_PHYSID1);
  if (phy_id == 0xffff)
    {
      printf("%s: No PHY present\n", netdev->name);
      return 0;
    }

  status = macb_mdio_read(macb, MII_BMSR);
  if (!(status & BMSR_LSTATUS))
    {
      /* Try to re-negotiate if we don't have link already. */
      macb_phy_reset(macb);

      for (i = 0; i < CONFIG_SYS_MACB_AUTONEG_TIMEOUT / 100; i++)
        {
          status = macb_mdio_read(macb, MII_BMSR);
          if (status & BMSR_LSTATUS)
            break;
          udelay(100);
        }
    }

  if (!(status & BMSR_LSTATUS))
    {
      printf("%s: link down (status: 0x%04x)\n",
             netdev->name, status);
      return 0;
    }
  else
    {
      adv = macb_mdio_read(macb, MII_ADVERTISE);
      lpa = macb_mdio_read(macb, MII_LPA);
      media = mii_nway_result(lpa & adv);
      speed = (media & (ADVERTISE_100FULL | ADVERTISE_100HALF)
               ? 1 : 0);
      duplex = (media & ADVERTISE_FULL) ? 1 : 0;
      printf("%s: link up, %sMbps %s-duplex (lpa: 0x%04x)\n",
             netdev->name,
             speed ? "100" : "10",
             duplex ? "full" : "half",
             lpa);

      ncfgr = macb_readl(macb, NCFGR);
      ncfgr &= ~(MACB_BIT(SPD) | MACB_BIT(FD));
      if (speed)
        ncfgr |= MACB_BIT(SPD);
      if (duplex)
        ncfgr |= MACB_BIT(FD);
      macb_writel(macb, NCFGR, ncfgr);
      return 1;
    }
}

#define ReadByte(addr)  *((volatile uint8_t*)(addr))
#define ReadHWord(addr) *((volatile uint16_t*)(addr))
#define ReadWord(addr)  *((volatile uint32_t*)(addr))

#define WriteByte(addr,value)  (*((volatile uint8_t*)(addr)) = (value))
#define WriteHWord(addr,value) (*((volatile uint16_t*)(addr)) = (value))
#define WriteWord(addr,value)  (*((volatile uint32_t*)(addr)) = (value))

#define NCS2  0x30000000
#define RWS_SetBit(reg,mask)  WriteHWord(NCS2+(reg), ReadHWord(NCS2+(reg)) | (mask))
#define RWS_ClrBit(reg,mask)  WriteHWord(NCS2+(reg), ReadHWord(NCS2+(reg)) & ~(mask))
#define RWS_Write(reg,val)    WriteHWord(NCS2+(reg), (val))
#define RWS_Read(reg)         ReadHWord(NCS2+(reg))

// Rockwell switch register offsets
#define TS_Control     0x00
#define TS_Event       0x02
#define Addend_Lo      0x04
#define Addend_Hi      0x06
#define Accum_Lo       0x08
#define Accum_Hi       0x0a
#define SysTime_Lo_Lo  0x0c
#define SysTime_Lo_Hi  0x0e
#define SysTime_Hi_Lo  0x10
#define SysTime_Hi_Hi  0x12
#define TgtTim1_Lo_Lo  0x14
#define TgtTim1_Lo_Hi  0x16
#define TgtTim1_Hi_Lo  0x18
#define TgtTim1_Hi_Hi  0x1a
#define TgtTim2_Lo_Lo  0x1c
#define TgtTim2_Lo_Hi  0x1e
#define TgtTim2_Hi_Lo  0x20
#define TgtTim2_Hi_Hi  0x22
#define Evnt1_Lo_Lo    0x24
#define Evnt1_Lo_Hi    0x26
#define Evnt1_Hi_Lo    0x28
#define Evnt1_Hi_Hi    0x2a
#define Evnt2_Lo_Lo    0x2c
#define Evnt2_Lo_Hi    0x2e
#define Evnt2_Hi_Lo    0x30
#define Evnt2_Hi_Hi    0x32
#define TxTS_Lo_Lo     0x34
#define TxTS_Lo_Hi     0x36
#define TxTS_Hi_Lo     0x38
#define TxTS_Hi_Hi     0x3a
#define RxTS_Lo_Lo     0x3c
#define RxTS_Lo_Hi     0x3e
#define RxTS_Hi_Lo     0x40
#define RxTS_Hi_Hi     0x42
#define RxID_Lo_Lo     0x44
#define RxID_Lo_Hi     0x46
#define RxID_Hi_Lo     0x48
#define RxID_Hi_Hi     0x4a
#define RxSeqID        0x4c
#define Subdomain      0x4e
#define SW_Control     0x50
#define SW_Event       0x52
#define SW_Mask        0x54
#define Prt1Seed       0x56
#define Prt2Seed       0x58
#define MAC_Lo_Lo      0x5a
#define MAC_Lo_Hi      0x5c
#define MAC_Hi_Lo      0x5e
#define MCHFI          0x60
#define MCHFV          0x62
#define FreeTmr_Lo     0x64
#define FreeTmr_Hi     0x66
#define PITTmo_Lo      0x68
#define PITTmo_Hi      0x6a
#define Rdndc_Control  0x6c
#define Rng_BcnState   0x6e
#define BcnTmo_Lo      0x70
#define BcnTmo_Hi      0x72
#define P1BcnTS_Lo     0x74
#define P1BcnTS_Hi     0x76
#define P2BcnTS_Lo     0x78
#define P2BcnTS_Hi     0x7a
#define P1NChkSId_Lo_Lo   0x7c
#define P1NChkSId_Lo_Hi   0x7e
#define P1NChkSId_Hi_Lo   0x80
#define P2NChkSId_Lo_Lo   0x82
#define P2NChkSId_Lo_Hi   0x84
#define P2NChkSId_Hi_Lo   0x86
#define P1NStsSId_Lo_Lo   0x88
#define P1NStsSId_Lo_Hi   0x8a
#define P1NStsSId_Hi_Lo   0x8c
#define P2NStsSId_Lo_Lo   0x8e
#define P2NStsSId_Lo_Hi   0x90
#define P2NStsSId_Hi_Lo   0x92
#define RngSprvsrId_Lo_Lo 0x94
#define RngSprvsrId_Lo_Hi 0x96
#define RngSprvsrId_Hi_Lo 0x98
#define Reserved9A      0x9a
#define P1RxGrpFrmCnt   0x9c
#define P2RxGrpFrmCnt   0x9e
#define P1RxUniFrmCnt   0xa0
#define P2RxUniFrmCnt   0xa2
#define P1RxBytCnt_Lo   0xa4
#define P1RxBytCnt_Hi   0xa6
#define P2RxBytCnt_Lo   0xa8
#define P2RxBytCnt_Hi   0xaa
#define P1RxLrgFrmCnt   0xac
#define P2RxLrgFrmCnt   0xae
#define P1RxAlnFrmCnt   0xb0
#define P2RxAlnFrmCnt   0xb2
#define P1RxFcsFrmCnt   0xb4
#define P2RxFcsFrmCnt   0xb6
#define P1RxErrFrmCnt   0xb8
#define P2RxErrFrmCnt   0xba
#define P1TxGrpFrmCnt   0xbc
#define P2TxGrpFrmCnt   0xbe
#define P1TxUniFrmCnt   0xc0
#define P2TxUniFrmCnt   0xc2
#define P1TxBytCnt_Lo   0xc4
#define P1TxBytCnt_Hi   0xc6
#define P2TxBytCnt_Lo   0xc8
#define P2TxBytCnt_Hi   0xca
#define P1TxSClFrmCnt   0xcc
#define P2TxSClFrmCnt   0xce
#define P1TxMClFrmCnt   0xd0
#define P2TxMClFrmCnt   0xd2
#define P1TxXClFrmCnt   0xd4
#define P2TxXClFrmCnt   0xd6
#define P1TxErrFrmCnt   0xd8
#define P2TxErrFrmCnt   0xda
#define BcnGenCntrl     0xdc
#define BcnIntrvl_Lo    0xde
#define BcnIntrvl_Hi    0xe0
#define BcnDataIdx      0xe2
#define BcnDataVal      0xe4
#define EmbSwtchVer     0xe6

#define RockwellPhy1Addr 4
#define RockwellPhy2Addr 2

#define Log() printf("%s: %d\n",__FUNCTION__,__LINE__)

static struct phyid_t
{
  u16 id[2];
  const char *name;
} supported_phy[] =
{
  {{0x0022, 0x1550}, "KSZ8051"},   // Micrel KSZ8051MNL
  {{0x0022, 0x1560}, "KSZ8081"}    // Micrel KSZ8081MLXIA
};

#define MII_STRAP_OVERRIDE 0x16  // Micrel strap-override register

// check to see if the PHY is supported
static int probe_phy(struct macb_device *macb, u16 addr)
{
  struct phyid_t phyid;
  int i;

  phyid.id[0] = mdio_read(macb, addr, MII_PHYSID1);
  phyid.id[1] = mdio_read(macb, addr, MII_PHYSID2);
  for (i = 0; i < (sizeof supported_phy)/(sizeof supported_phy[0]); ++i)
    if ((supported_phy[i].id[0] == phyid.id[0]) &&
        (supported_phy[i].id[1] == (phyid.id[1] & 0xfff0)))
      {
        printf("%s: found %s PHY at addr %d\n",macb->netdev.name,supported_phy[i].name,addr);
        mdio_write(macb, addr, MII_STRAP_OVERRIDE, 0x0001);
        return 1;
      }
  printf("%s: no PHY found at addr %d\n",macb->netdev.name, addr);
  return 0;
}

// check to see if rockwell switch is present
static void probe_rockwell_switch(struct macb_device *macb)
{
  unsigned short u;
  // first check for the two Micrel PHYs, since probing via MDIO
  // device ID read isn't going to cause any problems.

  if (!macb || macb->rockwell_switch_present)
    return;

  if (!probe_phy(macb, RockwellPhy1Addr) ||
      !probe_phy(macb, RockwellPhy2Addr))
    return;

  u = RWS_Read(MAC_Lo_Lo);

  // try to write/read a switch register
  RWS_Write(MAC_Lo_Lo,0xffff);
  if (RWS_Read(MAC_Lo_Lo) != 0xffff)
    return;
  RWS_Write(MAC_Lo_Lo,0x1234);
  if (RWS_Read(MAC_Lo_Lo) != 0x1234)
    return;
  RWS_Write(MAC_Lo_Lo,0x4321);
  if (RWS_Read(MAC_Lo_Lo) != 0x4321)
    return;
  RWS_Write(MAC_Lo_Lo,0x0000);
  if (RWS_Read(MAC_Lo_Lo) != 0x0000)
    return;

  RWS_Write(MAC_Lo_Lo,u);

  // switch version number should be present and not writable
  u = RWS_Read(EmbSwtchVer);
  if ((u & 0xfff0) != 0x0130)
    return;
  RWS_Write(EmbSwtchVer,0x0000);
  RWS_Write(EmbSwtchVer,0xffff);
  if (RWS_Read(EmbSwtchVer) != u)
    return;

  macb->rockwell_switch_present = 1;
  printf("%s: found Rockwell Ethernet Switch\n",macb->netdev.name);
}

#define SW_Event_Link1_Delta Bit(0)
#define SW_Event_Link1_Up    Bit(14)
#define SW_Event_Link2_Delta Bit(1)
#define SW_Event_Link2_Up    Bit(15)

#define SW_Control_Reset       Bit(0)
#define SW_Control_Port1_Reset Bit(1)
#define SW_Control_Port2_Reset Bit(2)
#define SW_Control_Port1_Half  Bit(3)
#define SW_Control_Port2_Half  Bit(4)
#define SW_Control_Port1_10M   Bit(5)
#define SW_Control_Port2_10M   Bit(6)
#define SW_Control_Port1_LedOn Bit(7)
#define SW_Control_Port2_LedOn Bit(8)
#define SW_Control_Port1_LedYellow Bit(9)
#define SW_Control_Port2_LedYellow Bit(10)

#define SW_Control_Unicast_Learning Bit(15)

// update the switch's external MAC configs if PHY link state has
// changed

static void linkup(uint16_t phystatus, unsigned port)
{
  RWS_SetBit(SW_Control, (SW_Control_Port1_Reset+SW_Control_Port1_LedOn)<<port);
  switch (phystatus & (Bit(1)+Bit(0)))
    {
    case Bit(0):
      printf("up 10M ");
      RWS_SetBit(SW_Control,SW_Control_Port1_10M<<port);
      break;
    case Bit(1):
      printf("up 100M ");
      RWS_ClrBit(SW_Control,SW_Control_Port1_10M<<port);
      break;
    default:
      printf("up ");
      break;
    }
  if (phystatus & Bit(2))
    {
      printf("full\n");
      RWS_ClrBit(SW_Control,SW_Control_Port1_Half<<port);
    }
  else
    {
      printf("half\n");
      RWS_SetBit(SW_Control,SW_Control_Port1_Half<<port);
    }
  RWS_ClrBit(SW_Control, SW_Control_Port1_Reset<<port);
}

static void linkdown(unsigned port)
{
  printf("down\n");
  RWS_ClrBit(SW_Control, SW_Control_Port1_LedOn<<port);
}

void update_rockwell_switch(void)
{
  unsigned short u,port;

  if (!macbp || !macbp->rockwell_switch_present || !macbp->rockwell_switch_running)
    return;

  u = RWS_Read(SW_Event);

  for (port=0; port<=1; ++port)
    if (u & (SW_Event_Link1_Delta<<port))
      {
        RWS_Write(SW_Event,(SW_Event_Link1_Delta<<port));  // clear it
        printf("port %d link ", port+1);
        if (u & (SW_Event_Link1_Up<<port))
          linkup(mdio_read(macbp, port ? RockwellPhy2Addr : RockwellPhy1Addr, 0x1e), port);
        else
          linkdown(port);
      }
}

void start_rockwell_switch(struct macb_device *macb)
{
  extern int eth_getenv_enetaddr(char *name, uchar *enetaddr);
  uchar mac[6];

  if (!macbp || macbp->rockwell_switch_running)
    return;

  eth_getenv_enetaddr("ethaddr",mac);

  printf("Rockwell Switch version: %04x\n", ReadHWord(NCS2+EmbSwtchVer));
  printf("Setting host MAC %02x:%02x:%02x:%02x:%02x:%02x\n",mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);

  // configure PHY LED0 pin to be link status signal for switch
  mdio_write(macb,RockwellPhy1Addr,0x1f,0x8110);
  mdio_write(macb,RockwellPhy2Addr,0x1f,0x8110);

  RWS_SetBit(SW_Control, SW_Control_Reset);

  // tell it the CPU's MAC address
  RWS_Write(MAC_Hi_Lo, (mac[0]<<8) | mac[1]);
  RWS_Write(MAC_Lo_Hi, (mac[2]<<8) | mac[3]);
  RWS_Write(MAC_Lo_Lo, (mac[4]<<8) | mac[5]);

  RWS_SetBit(SW_Control, SW_Control_Unicast_Learning);

  RWS_ClrBit(SW_Control, SW_Control_Port1_Reset+SW_Control_Port2_Reset);

  update_rockwell_switch();
  macbp->rockwell_switch_running = 1;
}


static int macb_init(struct eth_device *netdev, bd_t *bd)
{
  struct macb_device *macb = to_macb(netdev);
  unsigned long paddr;
  u32 hwaddr_bottom;
  u16 hwaddr_top;
  int i;

  /*
   * macb_halt should have been called at some point before now,
   * so we'll assume the controller is idle.
   */

  /* initialize DMA descriptors */
  paddr = macb->rx_buffer_dma;
  for (i = 0; i < CONFIG_SYS_MACB_RX_RING_SIZE; i++)
    {
      if (i == (CONFIG_SYS_MACB_RX_RING_SIZE - 1))
        paddr |= RXADDR_WRAP;
      macb->rx_ring[i].addr = paddr;
      macb->rx_ring[i].ctrl = 0;
      paddr += 128;
    }
  for (i = 0; i < CONFIG_SYS_MACB_TX_RING_SIZE; i++)
    {
      macb->tx_ring[i].addr = 0;
      if (i == (CONFIG_SYS_MACB_TX_RING_SIZE - 1))
        macb->tx_ring[i].ctrl = TXBUF_USED | TXBUF_WRAP;
      else
        macb->tx_ring[i].ctrl = TXBUF_USED;
    }
  macb->rx_tail = macb->tx_head = macb->tx_tail = 0;

  macb_writel(macb, RBQP, macb->rx_ring_dma);
  macb_writel(macb, TBQP, macb->tx_ring_dma);

  /* set hardware address */
  hwaddr_bottom = cpu_to_le32(*((u32 *)netdev->enetaddr));
  macb_writel(macb, SA1B, hwaddr_bottom);
  hwaddr_top = cpu_to_le16(*((u16 *)(netdev->enetaddr + 4)));
  macb_writel(macb, SA1T, hwaddr_top);

  /* choose RMII or MII mode. This depends on the board */
#ifdef CONFIG_RMII
#if defined(CONFIG_AT91CAP9) || defined(CONFIG_AT91SAM9260) || \
    defined(CONFIG_AT91SAM9263) || defined(CONFIG_AT91SAM9G20) || \
        defined(CONFIG_AT91SAM9G45) || defined(CONFIG_AT91SAM9M10G45)
  macb_writel(macb, USRIO, MACB_BIT(RMII) | MACB_BIT(CLKEN));
#else
  macb_writel(macb, USRIO, 0);
#endif
#else
#if defined(CONFIG_AT91CAP9) || defined(CONFIG_AT91SAM9260) || \
    defined(CONFIG_AT91SAM9263) || defined(CONFIG_AT91SAM9G20) || \
        defined(CONFIG_AT91SAM9G45) || defined(CONFIG_AT91SAM9M10G45)
  macb_writel(macb, USRIO, MACB_BIT(CLKEN));
#else
  macb_writel(macb, USRIO, MACB_BIT(MII));
#endif
#endif /* CONFIG_RMII */

  if (!macb_phy_init(macb))
    return -1;

  // check for presense of Rockwell Ethernet switch (but don't
  // actually do anything with it.  If switch is there, start it.

  probe_rockwell_switch(macb);
  if (macb->rockwell_switch_present)
    start_rockwell_switch(macb);

  /* Enable TX and RX */
  macb_writel(macb, NCR, MACB_BIT(TE) | MACB_BIT(RE));

  return 0;
}

static void macb_halt(struct eth_device *netdev)
{
  struct macb_device *macb = to_macb(netdev);
  u32 ncr, tsr;

  /* Halt the controller and wait for any ongoing transmission to end. */
  ncr = macb_readl(macb, NCR);
  ncr |= MACB_BIT(THALT);
  macb_writel(macb, NCR, ncr);

  do
    {
      tsr = macb_readl(macb, TSR);
    }
  while (tsr & MACB_BIT(TGO));

  /* Disable TX and RX, and clear statistics */
  macb_writel(macb, NCR, MACB_BIT(CLRSTAT));
}

int macb_eth_initialize(int id, void *regs, unsigned int phy_addr)
{
  struct macb_device *macb;
  struct eth_device *netdev;
  unsigned long macb_hz;
  u32 ncfgr;

  macb = malloc(sizeof(struct macb_device));
  if (!macb)
    {
      printf("Error: Failed to allocate memory for MACB%d\n", id);
      return -1;
    }

  macbp = macb;    // save device pointer for use by some command-line stuff

  memset(macb, 0, sizeof(struct macb_device));

  netdev = &macb->netdev;

  macb->rx_buffer = dma_alloc_coherent(CONFIG_SYS_MACB_RX_BUFFER_SIZE,
                                       &macb->rx_buffer_dma);
  macb->rx_ring = dma_alloc_coherent(CONFIG_SYS_MACB_RX_RING_SIZE
                                     * sizeof(struct macb_dma_desc),
                                     &macb->rx_ring_dma);
  macb->tx_ring = dma_alloc_coherent(CONFIG_SYS_MACB_TX_RING_SIZE
                                     * sizeof(struct macb_dma_desc),
                                     &macb->tx_ring_dma);

  macb->regs = regs;
  macb->phy_addr = phy_addr;

  sprintf(netdev->name, "macb%d", id);
  netdev->init = macb_init;
  netdev->halt = macb_halt;
  netdev->send = macb_send;
  netdev->recv = macb_recv;

  /*
   * Do some basic initialization so that we at least can talk
   * to the PHY
   */
  macb_hz = get_macb_pclk_rate(id);
  if (macb_hz < 20000000)
    ncfgr = MACB_BF(CLK, MACB_CLK_DIV8);
  else if (macb_hz < 40000000)
    ncfgr = MACB_BF(CLK, MACB_CLK_DIV16);
  else if (macb_hz < 80000000)
    ncfgr = MACB_BF(CLK, MACB_CLK_DIV32);
  else
    ncfgr = MACB_BF(CLK, MACB_CLK_DIV64);

  macb_writel(macb, NCFGR, ncfgr);

  eth_register(netdev);

#if defined(CONFIG_CMD_MII)
  miiphy_register(netdev->name, macb_miiphy_read, macb_miiphy_write);
#endif
  return 0;
}

#endif

static int do_startswitch(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  start_rockwell_switch(macbp);
  return 0;
}

U_BOOT_CMD(startswitch, 1, 0, do_startswitch, "start Rockwell switch", "");

static int do_switchtiming(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  extern void setSMC2timings(unsigned setup, unsigned pulse, unsigned hold);
  extern void showSMC2timings(void);
  unsigned setup, pulse, hold;
  if (argc == 1)
    {
      showSMC2timings();
      return 0;
    }

  if (argc != 4)
    {
      printf("<setup> <pulse> <hold> prameters required (in bus clocks)\n");
      return 1;
    }

  setup = simple_strtoul(argv[1], NULL, 10);
  pulse = simple_strtoul(argv[2], NULL, 10);
  hold = simple_strtoul(argv[3], NULL, 10);

  setSMC2timings(setup,pulse,hold);
  showSMC2timings();
  return 0;
}

U_BOOT_CMD(switchtiming, 4, 0, do_switchtiming, "set bus cycle times for Rockwell switch", "<setup> <pulse> <hold>");

int test5A(unsigned loop, unsigned short p)
{
  unsigned short s;
  RWS_Write(0x5A,p);
  s = RWS_Read(0x5A);
  if (0)
    if (s != p)
      {
        printf("%6u: exp %04x got %04x", loop, p, s);
        printf(" %04x", RWS_Read(0x5A));
        printf(" %04x\n", RWS_Read(0x5A));
      }
  return s!=p;
}

static int do_switchtest(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  unsigned short p __attribute__((unused));
  unsigned u, loops, failures, howmany = 0;

  if (argc == 2)
    howmany = simple_strtoul(argv[1], NULL, 10);

  printf("Test switch R/W register\n");

  printf("ffff/0000/5555/AAAA and walking 1/0\n");
  failures = 0;
  failures += test5A(0,0xffff);
  failures += test5A(0,0x0000);
  failures += test5A(0,0x5555);
  failures += test5A(0,0xAAAA);
  for (p=0x0001; p; p<<=1)
    {
      failures += test5A(0,p);
      failures += test5A(0,~p);
    }
  printf("             %u failures\n",failures);
  if (failures)
    return 1;

  printf("Incrementing pattern\n");

  failures = loops = 0;
  while (loops < 0xffffffff && (howmany == 0  || loops < howmany))
    {
      failures += test5A(u,(unsigned short)loops);
      ++loops;
      if ((loops % 100000) == 0)
        {
          if (tstc())
            break;
          if ((loops % 1000000) == 0)
            printf("%5uM loops %d failures\n",loops/1000000,failures);
        }
    }
  printf("Done\n%9u loops %u failures\n",loops,failures);

  return failures != 0;
}

U_BOOT_CMD(switchtest, 2, 0, do_switchtest, "Rockwell switch register R/W test", "[count]");

// Read low 16 bits first: for some counters this latches the high 16
// bits
unsigned RWS_Get32(unsigned reg)
{
  unsigned short lo = RWS_Read(reg);
  unsigned short hi = RWS_Read(reg+2);
  return (hi<<16) | lo;
}

// Some 32 bit counters don't latch the high 16 bits: you just keep
// trying until you get the same value twice in a row.
unsigned RWS_Get32m(unsigned reg)
{
  unsigned u;
  do
    u = RWS_Get32(reg);
  while (u != RWS_Get32(reg));
  return u;
}

static int do_switchcounters(cmd_tbl_t * cmdtp, int flag, int argc, char *argv[])
{
  printf("               %9s %9s\n","Port 1","Port 2");
  printf("Rx Bytes       %9u %9u\n",RWS_Get32m(P1RxBytCnt_Lo),RWS_Get32m(P2RxBytCnt_Lo));
  printf("Rx Unicst      %9u %9u\n",RWS_Read(P1RxUniFrmCnt),RWS_Read(P2RxUniFrmCnt));
  printf("Rx Group       %9u %9u\n",RWS_Read(P1RxGrpFrmCnt),RWS_Read(P2RxGrpFrmCnt));
  printf("Rx Err         %9u %9u\n",RWS_Read(P1RxErrFrmCnt),RWS_Read(P2RxErrFrmCnt));
  printf("Rx Err Large   %9u %9u\n",RWS_Read(P1RxLrgFrmCnt),RWS_Read(P2RxLrgFrmCnt));
  printf("Rx Err Align   %9u %9u\n",RWS_Read(P1RxAlnFrmCnt),RWS_Read(P2RxAlnFrmCnt));
  printf("Rx Err FCS     %9u %9u\n",RWS_Read(P1RxFcsFrmCnt),RWS_Read(P2RxFcsFrmCnt));
  printf("\n");
  printf("Tx Bytes       %9u %9u\n",RWS_Get32m(P1TxBytCnt_Lo),RWS_Get32m(P2TxBytCnt_Lo));
  printf("Tx Unicst      %9u %9u\n",RWS_Read(P1TxUniFrmCnt),RWS_Read(P2TxUniFrmCnt));
  printf("Tx Group       %9u %9u\n",RWS_Read(P1TxGrpFrmCnt),RWS_Read(P2TxGrpFrmCnt));
  printf("Tx Err         %9u %9u\n",RWS_Read(P1TxErrFrmCnt),RWS_Read(P2TxErrFrmCnt));
  printf("Tx Err SCl     %9u %9u\n",RWS_Read(P1TxSClFrmCnt),RWS_Read(P2TxSClFrmCnt));
  printf("Tx Err MCl     %9u %9u\n",RWS_Read(P1TxMClFrmCnt),RWS_Read(P2TxMClFrmCnt));
  printf("Tx Err XCl     %9u %9u\n",RWS_Read(P1TxXClFrmCnt),RWS_Read(P2TxXClFrmCnt));
  return 0;
}
U_BOOT_CMD(switchcounters, 1, 0, do_switchcounters, "Rockwell switch counter display", "");


// This file uses different indentation than the rest of the U-Boot
// source files, this is how let emacs know
//
// Local Variables:
// c-file-style: "gnu"
// End:
